/*
 * Copyright 2006 Google Inc.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.google.gwt.junit.rebind;

import com.google.gwt.core.ext.Generator;
import com.google.gwt.core.ext.GeneratorContext;
import com.google.gwt.core.ext.TreeLogger;
import com.google.gwt.core.ext.UnableToCompleteException;
import com.google.gwt.core.ext.typeinfo.JClassType;
import com.google.gwt.core.ext.typeinfo.JMethod;
import com.google.gwt.core.ext.typeinfo.NotFoundException;
import com.google.gwt.core.ext.typeinfo.TypeOracle;
import com.google.gwt.user.rebind.ClassSourceFileComposerFactory;
import com.google.gwt.user.rebind.SourceWriter;

import java.io.PrintWriter;
import java.util.HashSet;

/**
 * This class generates a stub class for classes that derive from GWTTestCase.
 * This stub class provides the necessary bridge between our Hosted or Hybrid
 * mode classes and the JUnit system.
 */
public class JUnitTestCaseStubGenerator extends Generator {
  private static final String GWT_TESTCASE_CLASS_NAME = "com.google.gwt.junit.client.GWTTestCase";

  private String reboundClassName;

  public JUnitTestCaseStubGenerator(String reboundClassName) {
    this.reboundClassName = reboundClassName;
  }

  /**
   * 
   */
  private SourceWriter getSourceWriter(TreeLogger logger, GeneratorContext ctx,
      String packageName, String className, String superclassName) {
    PrintWriter printWriter = ctx.tryCreate(logger, packageName, className);
    if (printWriter == null) {
      return null;
    }

    ClassSourceFileComposerFactory composerFactory = new ClassSourceFileComposerFactory(
      packageName, className);

    composerFactory.setSuperclass(superclassName);

    return composerFactory.createSourceWriter(ctx, printWriter);
  }

  /**
   * Create a new type that statisfies the rebind request.
   */
  public String generate(TreeLogger logger, GeneratorContext context,
      String typeName) throws UnableToCompleteException {

    TypeOracle typeOracle = context.getTypeOracle();
    assert typeOracle != null;

    if (!typeName.equals(GWT_TESTCASE_CLASS_NAME)) {
      logger.log(TreeLogger.ERROR, "Expected a rebind for class '"
        + GWT_TESTCASE_CLASS_NAME + "' but class '" + typeName
        + "' was requested instead", null);

      throw new UnableToCompleteException();
    }

    JClassType requestedClass;
    try {
      requestedClass = typeOracle.getType(reboundClassName);
    } catch (NotFoundException e) {
      logger.log(TreeLogger.ERROR, "Could not find type '" + typeName
        + "'; please see the log, as this usually indicates a previous error ",
        e);
      throw new UnableToCompleteException();
    }

    // Get the stub class name, and see if its source file exists.
    //
    String stubClassSimpleSourceName = getStubClassSimpleSourceName(requestedClass);

    String packageName = requestedClass.getPackage().getName();
    String qualifiedStubClassName = packageName + "."
      + stubClassSimpleSourceName;

    SourceWriter sw = getSourceWriter(logger, context, packageName,
      stubClassSimpleSourceName, requestedClass.getQualifiedSourceName());
    if (sw == null) {
      return qualifiedStubClassName;
    }

    // writeWindowExternalAccessors(sw);

    String[] testMethods = getTestMethodNames(requestedClass);
    writeOnModuleLoad(sw);
    writeExecuteTestMethod(testMethods, sw);
    writeReportError(sw);

    sw.commit(logger);

    return qualifiedStubClassName;
  }

  /**
   * Given a class return all methods that are considered JUnit test methods up
   * to but not including the declared methods of the class named
   * GWT_TESTCASE_CLASS_NAME.
   */
  private String[] getTestMethodNames(JClassType requestedClass) {
    HashSet testMethodNames = new HashSet();
    JClassType cls = requestedClass;

    while (true) {
      // We do not consider methods in the GWT superclass or above
      //
      if (isGWTTestCaseClass(cls)) {
        break;
      }

      JMethod[] clsDeclMethods = cls.getMethods();
      for (int i = 0, n = clsDeclMethods.length; i < n; ++i) {
        JMethod declMethod = clsDeclMethods[i];

        // Skip methods that are not JUnit test methods.
        //
        if (!isJUnitTestMethod(declMethod)) {
          continue;
        }

        if (testMethodNames.contains(declMethod.getName())) {
          continue;
        }

        testMethodNames.add(declMethod.getName());
      }

      cls = cls.getSuperclass();
    }

    return (String[]) testMethodNames
      .toArray(new String[testMethodNames.size()]);
  }

  /**
   * Returns true if the method is considered to be a valid JUnit test method.
   * The criteria are that the method's name begin with "test", have public
   * access, and not be static.
   */
  private boolean isJUnitTestMethod(JMethod method) {
    if (!method.getName().startsWith("test")) {
      return false;
    }

    if (!method.isPublic() || method.isStatic()) {
      return false;
    }

    return true;
  }

  /**
   * Returns true if the class is the special GWT Test Case derived class
   */
  private boolean isGWTTestCaseClass(JClassType cls) {
    return cls.getQualifiedSourceName()
      .equalsIgnoreCase(GWT_TESTCASE_CLASS_NAME);
  }

  /**
   * Create the appMain method that is the main entry point for the GWT
   * application.
   */
  private void writeOnModuleLoad(SourceWriter sw) {
    sw.println();
    sw.println("public final void onModuleLoad() {");
    sw.indent();
    sw
      .println("com.google.gwt.user.client.DeferredCommand.add(new com.google.gwt.user.client.Command() {");
    sw.indent();
    sw.println("public void execute() {");
    sw.indent();

    // If the test case wants exceptions to be caught by the browser.
    //
    sw.println("if (!catchExceptions()) {");
    sw.indent();
    sw.println("executeTestMethod();");
    sw.outdent();
    sw.println("} else {");

    // Otherwise, use the normal JUnit catch behavior.
    //
    sw.indent();
    sw.println("try {");
    sw.indent();
    sw.println("executeTestMethod();");
    sw.outdent();
    sw.println("} catch (junit.framework.AssertionFailedError e) {");
    sw
      .indentln("addFailure(e.getMessage(), com.google.gwt.core.client.GWT.isScript() ? null : e.getStackTrace());");
    sw.println("} catch (Throwable e) {");
    sw
      .indentln("reportError(e);");
    sw.println("}");
    sw.outdent();
    sw.println("}");
    sw.println("finishTestMethod();");
    sw.outdent();
    sw.println("}"); // finish execute()
    sw.outdent();
    sw.println("});"); // finish add()
    sw.outdent();
    sw.println("}"); // finish onModuleLoad();
  }

  private void writeExecuteTestMethod(String[] testMethodNames, SourceWriter sw) {
    sw.println();
    sw.println("public final void executeTestMethod() {");
    sw.indent();
    sw.println("String testName = getTestName();");
    for (int i = 0, n = testMethodNames.length; i < n; ++i) {
      String methodName = testMethodNames[i];

      if (i > 0) {
        sw.print("else ");
      }

      sw.println("if (testName.equals(\"" + methodName + "\")) {");
      sw.indentln(methodName + "();");
      sw.println("}");
    }
    sw.outdent();
    sw.println("}");
  }
  
  private void writeReportError(SourceWriter sw) {
    sw.println();
    sw.println("private void reportError(Throwable e) {");
    sw.indent();
    sw.println("if (com.google.gwt.core.client.GWT.isScript()) {");
    sw.indent();
    sw.println("String msg = e.toString();");
    sw.println("addError(null, msg);");
    sw.outdent();
    sw.println("} else {");
    sw.indent();
    sw.println("addError(e, null);");
    sw.outdent();
    sw.println("}");
    sw.outdent();
    sw.println("}");
  }
  
  /**
   * Gets the name of the native stub class.
   */
  private String getStubClassSimpleSourceName(JClassType baseClass) {
    return "__" + baseClass.getSimpleSourceName() + "_unitTestImpl";
  }
}
